#!/opt/edge/venv/bin/python3.11

import sys

from typing import Final, Self
from functools import reduce

from edge import eventlog
from edge.types.exceptions import *
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.api_terminal import TerminalApi
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uimod import uitype


class _CliConf:
    ACTION_CREATE = 'create'
    ACTION_CREATE_WAGA = 'create-waga'
    ACTION_CREATE_WSPASES_FLEX_R3D = 'create-wspases-flex-r3d'
    ACTION_REMOVE = 'remove'
    ACTION_REGENERATE = 'regenerate'

    _CREATE_ACTION_SPOOL_SIZE = '--spool-size'
    _CREATE_ACTION_VENDOR = '--vendor'
    _CREATE_ACTION_MODEL = '--model'
    _CREATE_ACTION_SERIAL = '--serial'
    _CREATE_ACTION_UUID = '--uuid'

    _REMOVE_ACTION_PID = '--pid'

    _ARG_VALUE_SEPARATOR = '='

    def __init__(
            self,
            action: str,
            create_vendor: str | None = None,
            create_model: str | None = None,
            create_serial: str | None = None,
            remove_pid: str | None = None,
            create_uuid: str | None = None,
            spool_size: str | None = None):

        self.action = action
        self.create_vendor = create_vendor
        self.create_model = create_model
        self.create_serial = create_serial
        self.create_uuid = create_uuid
        self.remove_pid = remove_pid
        self.create_spool_size = spool_size

    @classmethod
    def show_usage(cls) -> None:
        print(('pipeline-ctl create-waga [{}{}<spool-size>] [{}{}<uuid>]'
               .format(cls._CREATE_ACTION_SPOOL_SIZE, cls._ARG_VALUE_SEPARATOR,
                       cls._CREATE_ACTION_UUID, cls._ARG_VALUE_SEPARATOR)))

        print(('pipeline-ctl create-wspases-flex-r3d [{}{}<spool-size>] [{}{}<uuid>]'
               .format(cls._CREATE_ACTION_SPOOL_SIZE, cls._ARG_VALUE_SEPARATOR,
                       cls._CREATE_ACTION_UUID, cls._ARG_VALUE_SEPARATOR)))

        print(('pipeline-ctl create {}{}<vendor> {}{}<model> {}{}<serial> [{}{}<spool-size>] [{}{}<uuid>]'
               .format(cls._CREATE_ACTION_VENDOR, cls._ARG_VALUE_SEPARATOR,
                       cls._CREATE_ACTION_MODEL, cls._ARG_VALUE_SEPARATOR,
                       cls._CREATE_ACTION_SERIAL, cls._ARG_VALUE_SEPARATOR,
                       cls._CREATE_ACTION_SPOOL_SIZE, cls._ARG_VALUE_SEPARATOR,
                       cls._CREATE_ACTION_UUID, cls._ARG_VALUE_SEPARATOR)))

        print(('pipeline-ctl remove {}{}<pipeline-id>'
               .format(cls._REMOVE_ACTION_PID, cls._ARG_VALUE_SEPARATOR)))

        print('pipeline-ctl regenerate')

    @classmethod
    def _deserialize_create(cls, argv: list[str]) -> Self:
        if len(argv) < 3:
            raise ValueError('invalid number of arguments')

        vendor = None
        model = None
        serial = None
        new_uuid = None
        spool_size = None
        for arg in argv:
            raw_name_value = arg.split(cls._ARG_VALUE_SEPARATOR, 1)
            if len(raw_name_value) != 2:
                raise ValueError('invalid argument %s' % arg)

            raw_name, raw_value = raw_name_value
            raw_name = raw_name.strip()
            raw_value = raw_value.strip()

            if raw_name == cls._CREATE_ACTION_VENDOR:
                vendor = raw_value
                continue

            if raw_name == cls._CREATE_ACTION_SERIAL:
                serial = raw_value
                continue

            if raw_name == cls._CREATE_ACTION_MODEL:
                model = raw_value
                continue

            if raw_name == cls._CREATE_ACTION_SPOOL_SIZE:
                spool_size = raw_value
                continue

            if raw_name == cls._CREATE_ACTION_UUID:
                import uuid
                new_uuid = raw_value
                uuid.UUID(new_uuid)
                continue

            raise ValueError('invalid argument %s' % arg)

        if not reduce(lambda x, y: x and y, [vendor, model, serial]):
            raise ValueError('invalid number of arguments')

        return cls(cls.ACTION_CREATE,
                   create_vendor=vendor,
                   create_model=model,
                   create_serial=serial,
                   create_uuid=new_uuid,
                   spool_size=spool_size)

    @classmethod
    def _deserialize_remove(cls, argv: list[str]) -> Self:
        if len(argv) != 1:
            raise ValueError('invalid number of arguments')

        raw_name_value = str(argv[0]).split(cls._ARG_VALUE_SEPARATOR, 1)
        if len(raw_name_value) != 2:
            raise ValueError('invalid argument')

        raw_name, raw_value = raw_name_value
        raw_name = raw_name.strip()
        raw_value = raw_value.strip()

        if raw_name != cls._REMOVE_ACTION_PID:
            raise ValueError('invalid argument')

        return cls(cls.ACTION_REMOVE, remove_pid=raw_value)

    @classmethod
    def _deserialize_regenerate(cls, argv: list[str]) -> Self:
        if len(argv) != 0:
            raise ValueError('invalid number of arguments')

        return cls(cls.ACTION_REGENERATE)

    @classmethod
    def _deserialize_create_waga(cls, argv: list[str]) -> Self:
        if len(argv) not in [0, 1, 2]:
            raise ValueError('invalid number of arguments')

        new_uuid = None
        spool_size = None
        for arg in argv:
            raw_name_value = arg.split(cls._ARG_VALUE_SEPARATOR, 1)
            if len(raw_name_value) != 2:
                raise ValueError('invalid argument %s' % arg)

            raw_name, raw_value = raw_name_value
            raw_name = raw_name.strip()
            raw_value = raw_value.strip()

            if raw_name == cls._CREATE_ACTION_UUID:
                import uuid
                new_uuid = raw_value
                uuid.UUID(new_uuid)
                continue

            if raw_name == cls._CREATE_ACTION_SPOOL_SIZE:
                spool_size = raw_value
                continue

            raise ValueError('invalid argument %s' % arg)

        return cls(cls.ACTION_CREATE_WAGA, create_uuid=new_uuid, spool_size=spool_size)

    @classmethod
    def _deserialize_create_wspases_flex_r3d(cls, argv: list[str]) -> Self:
        if len(argv) not in [0, 1, 2]:
            raise ValueError('invalid number of arguments')

        new_uuid = None
        spool_size = None
        for arg in argv:
            raw_name_value = arg.split(cls._ARG_VALUE_SEPARATOR, 1)
            if len(raw_name_value) != 2:
                raise ValueError('invalid argument %s' % arg)

            raw_name, raw_value = raw_name_value
            raw_name = raw_name.strip()
            raw_value = raw_value.strip()

            if raw_name == cls._CREATE_ACTION_UUID:
                import uuid
                new_uuid = raw_value
                uuid.UUID(new_uuid)
                continue

            if raw_name == cls._CREATE_ACTION_SPOOL_SIZE:
                spool_size = raw_value
                continue

            raise ValueError('invalid argument %s' % arg)

        return cls(cls.ACTION_CREATE_WSPASES_FLEX_R3D, create_uuid=new_uuid, spool_size=spool_size)

    @classmethod
    def deserialize(cls, argv: list[str]) -> Self:
        arg_count = len(argv)
        if arg_count < 2:
            raise ValueError('invalid number of arguments')

        action = str(argv[1])

        if action == cls.ACTION_CREATE:
            return cls._deserialize_create(argv[2:])

        elif action == cls.ACTION_REMOVE:
            return cls._deserialize_remove(argv[2:])

        elif action == cls.ACTION_REGENERATE:
            return cls._deserialize_regenerate(argv[2:])

        elif action == cls.ACTION_CREATE_WAGA:
            return cls._deserialize_create_waga(argv[2:])

        elif action == cls.ACTION_CREATE_WSPASES_FLEX_R3D:
            return cls._deserialize_create_wspases_flex_r3d(argv[2:])

        raise ValueError('invalid command-line')

def _handle_regenerate_bs64k() -> None:
    from edge.webserver.sysconf.edge_creator import generate
    from edge.webserver.sysconf.uimod.eventstore import EsUimod

    root_info = ph.RootInfoUiconf.get_info()
    eventstores_info = ph.InfoUiconf.map_units_r(root_info, unit_types=[uitype.UI_TYPE_EVENTSTORE])

    for eventstore_info in eventstores_info:
        eventstore_ref = eventstore_info['ref']
        eventstore_uimod = EsUimod.create(eventstore_ref)
        try:
            # create edge-jnode-set
            # affected_uirefs = ph.InfoUiconf.find_unit_refs_r(eventstore_info)
            affected_uirefs = [eventstore_ref]
            generate(eventstore_uimod.create_backup_dirpath(), affected_uirefs, forced=True)

            eventlog.log_d('done: {}'.format(eventstore_ref))

        except BusyError:
            eventlog.log_d('fail: {} (lock error)'.format(eventstore_ref))

        except Exception:
            eventlog.log_d('fail: {} (traceback)'.format(eventstore_ref))
            eventlog.log_x()

        return


    eventlog.log_d('done')


def _handle_regenerate() -> None:
    from edge.webserver.sysconf.edge_creator import generate

    root_info = ph.RootInfoUiconf.get_info()
    terms_info = ph.InfoUiconf.map_units_r(root_info, unit_types=[uitype.UI_TYPE_TERMINAL])

    for term_info in terms_info:
        term_uimod = ph.UiMod.create(term_info['ref'])
        eventlog.log_d('exec: {}'.format(term_uimod.ref))
        # noinspection PyBroadException
        try:
            # create edge-jnode-set
            affected_uirefs = ph.InfoUiconf.find_unit_refs_r(term_info)
            factory_uimod = ph.UiMod(term_uimod.pid, uitype.UI_TYPE_FACTORY, '0')
            generate(factory_uimod.create_backup_dirpath(), affected_uirefs, forced=True)

            eventlog.log_d('done: {}'.format(term_uimod.ref))

        except BusyError:
            eventlog.log_d('fail: {} (lock error)'.format(term_uimod.ref))

        except Exception:
            eventlog.log_d('fail: {} (traceback)'.format(term_uimod.ref))
            eventlog.log_x()

    eventlog.log_d('done')


def _main(argv: list[str]) -> None:
    try:
        conf: Final[_CliConf] = _CliConf.deserialize(argv)
    except ValueError:
        eventlog.log_x()
        _CliConf.show_usage()
        exit(1)

    if conf.action == _CliConf.ACTION_CREATE:
        root_info = ph.RootInfoUiconf.get_info()
        ctrl_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_type=uitype.UI_TYPE_CONTROLLER)
        TerminalApi.create_term(ctrl_info['ref'], {
            'vendor': conf.create_vendor,
            'model': conf.create_model,
            'serial': conf.create_serial,
            'internal': True,
            'uuid': conf.create_uuid,
            'spool-size': conf.create_spool_size,
            'factory': 'radar-terminal-info.json'
        })

    elif conf.action == _CliConf.ACTION_CREATE_WAGA:
        root_info = ph.RootInfoUiconf.get_info()
        ctrl_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_type=uitype.UI_TYPE_CONTROLLER)
        TerminalApi.create_term(ctrl_info['ref'], {
            'internal': True,
            'uuid': conf.create_uuid,
            'with-intrinsic': True,
            'spool-size': conf.create_spool_size,
            'factory': 'rtsp-waga-terminal-info.json'
        })

    elif conf.action == _CliConf.ACTION_CREATE_WSPASES_FLEX_R3D:
        root_info = ph.RootInfoUiconf.get_info()
        ctrl_info = ph.RootInfoUiconf.find_unit_r(root_info, unit_type=uitype.UI_TYPE_CONTROLLER)
        TerminalApi.create_term(ctrl_info['ref'], {
            'internal': True,
            'uuid': conf.create_uuid,
            'with-intrinsic': True,
            'spool-size': conf.create_spool_size,
            'factory': 'wspases-flex-r3d-terminal-info.json'
        })

    elif conf.action == _CliConf.ACTION_REMOVE:
        TerminalApi.drop_terminal2(conf.remove_pid)
        dg.LifeCycle.sighup_ss_dgmod()
        dg.LifeCycle.restart_ws_dgmod()

    elif conf.action == _CliConf.ACTION_REGENERATE:
        _handle_regenerate()
        _handle_regenerate_bs64k()

    exit(0)


if __name__ == '__main__':
    eventlog.open_log('pipeline-ctl', 'debug')
    # noinspection PyBroadException
    try:
        _main(sys.argv)

    except Exception:
        eventlog.log_x()
        exit(1)

    finally:
        eventlog.close_log()
