import os

from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uimod import uitype
from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgstats.blobstore import BsDgstats
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod.eventstore import EsUimod
import edge.webserver.sysconf.edge_creator as creator


def _get_disk_size() -> float:
    result = os.statvfs(ph.CTRL_ARCH_DIR)
    block_size = result.f_frsize
    total_blocks = result.f_blocks

    mega = 1024 * 1024
    total_size = total_blocks * block_size / mega

    return total_size


# noinspection PyClassHasNoInit
class EventstoreApi:
    @classmethod
    def _get_limits(cls) -> dict:
        total_size = _get_disk_size()

        limits = {
            'cleanup': {
                'min-df': {
                    'min': 0,
                    'max': total_size
                },
                'max-du': {
                    'min': 0,
                    'max': total_size
                },
                'max-abs-time': {
                    'min': 0,
                    'max': 4294967295
                },
                'max-rel-time': {
                    'min': 0,
                    'max': 4294967295
                }
            }
        }
        return limits

    @classmethod
    def load_unit(cls, uimod_ref: str) -> dict:
        uimod = EsUimod.create(uimod_ref)
        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        stats_tz = TimeUiconf.get_tz()
        stats = BsDgstats.get_chan_stats_64k(uimod_ref.split('.')[-1], stats_tz)

        ext_uiconf_pub: dict = uiconf['rw']
        reply = {
            'type': uitype.UI_TYPE_EVENTSTORE,
            'conf': ext_uiconf_pub['args'],
            'limits': cls._get_limits()
        }
        if stats:
            reply['stats'] = stats

        return reply

    @classmethod
    def save_unit(cls, uimod_ref: str, new_uiconf: dict) -> dict:
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = EsUimod.create(uimod_ref)
        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        fs.validate_json({'args': new_uiconf_conf}, uimod.schema_path)

        uiconf_pub: dict = uiconf['rw']
        if uiconf_pub['args'] !=  new_uiconf_conf:
            if new_uiconf_conf['name'] != uiconf_pub['args']['name']:
                ph.RootInfoUiconf.set_unit_name(uimod_ref, new_uiconf_conf['name'])

            uiconf_pub['args'] = new_uiconf_conf

            # uimod.save_conf(uiconf)
            # bs.bs65535_reappend_chan(uiconf)
            uirefs = [uimod.ref]
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod.ref: uiconf})
            dg.CTRL_BS_DGMOD.sighup()

        stats_tz = TimeUiconf.get_tz()
        stats = BsDgstats.get_chan_stats_64k(uimod_ref.split('.')[-1], stats_tz)

        ext_uiconf_pub = uiconf_pub
        reply = {
            'type': uitype.UI_TYPE_EVENTSTORE,
            'conf': ext_uiconf_pub['args'],
            'limits': cls._get_limits()
        }
        if stats:
            reply['stats'] = stats

        return reply

    @classmethod
    def make_unit(cls, factory_conf: dict) -> dict:
        root_info, root_links = ph.RootInfoUiconf.get_info_links()
        ctrl_info = ph.InfoUiconf.find_unit_r(root_info, unit_type=uitype.UI_TYPE_CONTROLLER)
        # TODO es_uimod_id
        _, _, es_uimod_id = ph.RootInfoUiconf.new_uimod_pid_type_id(ctrl_info,
                                                                    uitype.UI_TYPE_EVENTSTORE)

        # add store_jnode to /etc/edge/setupstore/admin/action/eventstore/<chan-ref>/0.json
        es_uimod_ref = '{}.{}'.format(uitype.UI_TYPE_EVENTSTORE, es_uimod_id)
        uimod = EsUimod.create(es_uimod_ref)
        uiconf = uimod.install(factory_conf['name'], es_uimod_id)

        # bs.bs65535_append_chan(uiconf)
        uirefs = [uimod.ref]
        creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod.ref: uiconf})
        dg.CTRL_BS_DGMOD.sigusr1()

        ctrl_units: list = ctrl_info['units']
        ctrl_units.append({
            'internal': False,
            'role': uitype.UI_ROLE_ARCHIVE,
            'type': uitype.UI_TYPE_EVENTSTORE,
            'name': factory_conf['name'],
            'ref': es_uimod_ref
        })

        ph.RootInfoUiconf.set_info_links(root_info, root_links)

        root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        return {'info': [root_info], 'created': [es_uimod_ref]}

    @classmethod
    def delete_unit(cls, uimod_ref: str) -> dict:
        uimod = EsUimod.create(uimod_ref)

        root_info, changed_uirefs = ph.RootInfoUiconf.del_unit_r(uimod_ref)

        creator.generate(uimod.create_backup_dirpath(), changed_uirefs)
        dg.CTRL_BS_DGMOD.sigusr1()

        uimod.uninstall()
        # bs.bs65535_remove_chan(uimod_ref)
        reply = {
            'info': [root_info]
        }
        return reply
