from edge.webserver.sysconf.uimod import uitype
from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import ph

import edge.webserver.sysconf.edge_creator as creator


def get_limits() -> dict:

    jnode = {
        'acquisition': {
            'accumulations': [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024],
            'oversampling': [1, 2, 4, 8],
            'point-count': {
                'min': 2,
                'max': 64
            },
            'threshold-offset': {
                'min': -5.0,
                'max': 100.0
            },
            'smoothing': {
                'min': -17,
                'max': 16
            }
        }
    }

    return jnode


class LeddarApi:
    def __init__(self):
        pass

    @staticmethod
    def make_service(term_pid: str, factory_conf: dict) -> tuple[str, dict]:
        if not ph.RootInfoUiconf.validate_term(term_pid):
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        service_uimod_pid = term_pid
        service_uimod_type = factory_conf.get('type')
        service_name = factory_conf.get('name')
        factory_conf_conf = factory_conf.get('conf', {})

        conn_conf = factory_conf_conf.get('connection', {})
        conn_type = conn_conf.get('type', 'usb')
        conn_endpoint = None
        conn_address = 1
        if conn_type == 'modbus':
            conn_address = conn_conf.get('address', 1)
            conn_endpoint = conn_conf.get('endpoint')
            if conn_endpoint is None:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError
        elif conn_type != 'usb':
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        if service_uimod_type != uitype.UI_TYPE_LEDDAR or service_name is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        answer = ph.RootInfoUiconf.add_service(
            service_uimod_pid, service_uimod_type, service_name, uitype.UI_TYPE_TERMINAL,
            [uitype.UI_TYPE_TRAFFIC_RULES, uitype.UI_TYPE_LPR])

        if answer is None:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        service_uimod_id, root_info = answer

        uimod = ph.UiMod(service_uimod_pid, service_uimod_type, service_uimod_id)
        uimod.install()
        uiconf = uimod.load_conf()
        uiconf['rw']['conf']['name'] = service_name
        uiconf['ro']['connection'] = {
            'type': conn_type,
            'endpoint': conn_endpoint,
            'address': conn_address
        }

        uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod.ref: uiconf})
        return uimod.ref, root_info

    @staticmethod
    def drop_service(uimod_ref: str) -> dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(uitype.UI_TYPE_LEDDAR)
        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        root_info, changed_uirefs = ph.RootInfoUiconf.del_unit_r(uimod.ref)
        creator.generate(uimod.create_backup_dirpath(), changed_uirefs)
        uimod.uninstall()

        ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
        reply = {
            'info': [ext_root_info]
        }
        return reply

    @staticmethod
    def load_service(uimod_ref: str) -> dict:
        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(uitype.UI_TYPE_LEDDAR)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_e('fail: NotFoundError')
            raise NotFoundError

        uiconf_pub = uiconf['rw']

        reply = {
            'type': uitype.UI_TYPE_LEDDAR,
            'conf': uiconf_pub['conf'],
            'limits': get_limits()
        }
        return reply

    @staticmethod
    def save_service(uimod_ref: str, new_uiconf: dict) -> dict:
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(uimod_ref)
        uimod.assert_type(uitype.UI_TYPE_LEDDAR)

        if uimod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        def validate_body():
            fs.validate_json({'conf': new_uiconf_conf}, uimod.schema_path)

        validate_body()

        uiconf = uimod.load_conf()
        if uiconf['rw']['conf'] != new_uiconf_conf:
            if new_uiconf_conf['name'] != uiconf['rw']['conf']['name']:
                ph.RootInfoUiconf.set_unit_name(uimod.ref, new_uiconf_conf['name'])

            uiconf['rw']['conf'] = new_uiconf_conf
            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod.ref: uiconf})

        reply = {
            'type': uitype.UI_TYPE_LEDDAR,
            'conf': new_uiconf_conf,
            'limits': get_limits()
        }

        return reply
