import uuid

from typing import Callable

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import ph, edge_creator as creator
from edge.webserver.sysconf.api_leddar import LeddarApi
from edge.webserver.sysconf.api_seatbelt import SeatbeltApi
from edge.webserver.sysconf.api_tcp_recipient import TcpRecipientApi
from edge.webserver.sysconf.dgconf.setupstore import SsDgconf
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod import uitype


# noinspection PyClassHasNoInit
class TerminalApi:
    @classmethod
    def create_term(cls, ctrl_uiref: str, factory_conf: dict) -> dict:
        factory_name = factory_conf['factory']
        factory_uiconf_path = fs.join_path(ph.CTRL_DEF_UICONF_DIR, factory_name)
        factory_info, factory_links = ph.InfoUiconf.get_info_links(factory_uiconf_path)

        internal = factory_conf.get('internal', False)
        root_info, term_info = ph.RootInfoUiconf.add_term(ctrl_uiref, factory_info, factory_links, internal)

        term_ref = term_info['ref']
        term_pid = ph.UiMod.create(term_ref).pid
        term_uuid = factory_conf.get('uuid') or str(uuid.uuid4())

        factory_conf['term-pid'] = term_pid
        factory_conf['term-uuid'] = term_uuid

        def install_unit_fn(unit_info: dict) -> None:
            unit_uimod = ph.UiMod.create(unit_info['ref'])
            unit_uimod.install(factory_conf)

        # noinspection PyBroadException
        try:
            # install uimods
            TimeUiconf.install()
            factory_uimod = ph.UiMod(term_pid, uitype.UI_TYPE_FACTORY, '0')
            ph.InfoUiconf.map_units_r(term_info, install_unit_fn)
            ph.TermAboutUiconf.save_conf(term_pid, term_uuid)

            # install dgmods
            uirefs = ph.InfoUiconf.find_unit_refs_r(term_info)
            uirefs.append('webserver.0')  # edit ws conf while addind pipeline
            all_used_dgmods = creator.generate(factory_uimod.create_backup_dirpath(), uirefs, forced=True)

            all_used_dgmods = [
                x for x in all_used_dgmods if x.ref not in (dg.CTRL_WS_DGMOD.ref, dg.CTRL_ES_DGMOD.ref)
            ]

            spool_size = factory_conf.get('spool-size') or 740
            SsDgconf.append_ss_group(term_pid, [x.conf_dir for x in all_used_dgmods], int(spool_size))

            from edge.support import db
            db.edit_term(term_uuid, term_name=term_pid, term_place=term_pid, manual=True)

            dg.LifeCycle.restart_ws_dgmod()
            dg.LifeCycle.sighup_ss_dgmod()

            ext_root_info = ph.RootInfoUiconf.get_ext_root_info(root_info)
            return {'info': [ext_root_info], 'created': [term_ref]}

        except NotImplementedError:
            eventlog.log_d('fail: ParamsError (unsupported camera)')
            raise ParamsError

        except BusyError:
            raise

        except Exception:
            eventlog.log_x()
            TerminalApi.drop_terminal2(term_pid)
            raise InternalError

    @staticmethod
    def make_unit(term_uiref: str, factory_conf: dict) -> dict:
        uimod_type = factory_conf.get('type')
        if uimod_type is None:
            eventlog.log_d('fail: ParamsError (type)')
            raise ParamsError

        from .api_parking import ParkingApi
        from .api_wrong_way import WrongWayApi
        from .api_bus_lane import BusLaneApi
        from .api_red_light import RedLightApi
        from .api_wanted import WantedApi
        from .api_section_speed import SectionSpeedApi
        from .api_spot_speed import SpotSpeedApi
        from .api_railroad_crossing import RailroadCrossingApi

        uitype_to_factory_map = {
            uitype.UI_TYPE_DET_PARKING: ParkingApi.make_detector,
            uitype.UI_TYPE_DET_REDLIGHT: RedLightApi.make_detector,
            uitype.UI_TYPE_DET_SECTION_SPEED: SectionSpeedApi.make_detector,
            uitype.UI_TYPE_DET_SPOT_SPEED: SpotSpeedApi.make_detector,
            uitype.UI_TYPE_DET_WANTED: WantedApi.make_detector,
            uitype.UI_TYPE_DET_WRONG_WAY: WrongWayApi.make_detector,
            uitype.UI_TYPE_DET_BUS_LANE: BusLaneApi.make_detector,
            uitype.UI_TYPE_TCP_RECIPIENT: TcpRecipientApi.make_service,
            uitype.UI_TYPE_LEDDAR: LeddarApi.make_service,
            uitype.UI_TYPE_DET_RAILROAD: RailroadCrossingApi.make_detector,
            uitype.UI_TYPE_DET_SEATBELT: SeatbeltApi.make_detector
        }

        factory_fn = uitype_to_factory_map.get(uimod_type)
        if not factory_fn:
            eventlog.log_d('fail: ParamsError (type)')
            raise ParamsError

        term_uimod = ph.UiMod.create(term_uiref)
        unit_uiref, terms_info = factory_fn(term_uimod.pid, factory_conf)
        ext_terms_info = ph.RootInfoUiconf.get_ext_root_info(terms_info)

        reply = {
            'info': [ext_terms_info],
            'created': [unit_uiref]
        }
        return reply

    @staticmethod
    def get_factory_limits(term_uiref: str, roles: list[str]) -> list[dict]:
        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(uitype.UI_TYPE_TERMINAL)

        service_list = [
            {'role': uitype.UI_ROLE_SERVICE, 'type': uitype.UI_TYPE_TCP_RECIPIENT},
            {'role': uitype.UI_ROLE_SERVICE, 'type': uitype.UI_TYPE_LEDDAR}
        ]

        detector_list = [
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_REDLIGHT},
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_RAILROAD},
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_WANTED},
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_WRONG_WAY},
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_BUS_LANE},
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_SECTION_SPEED},
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_PARKING},
            {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_SEATBELT}
        ]

        if ph.RootInfoUiconf.get_raterm_pid() == uimod.pid:
            detector_list.append(
                {'role': uitype.UI_ROLE_DETECTOR, 'type': uitype.UI_TYPE_DET_SPOT_SPEED})

        if not roles:
            roles = [uitype.UI_ROLE_DETECTOR,
                     uitype.UI_ROLE_SERVICE]

        reply = []
        if uitype.UI_ROLE_DETECTOR in roles:
            reply.extend(detector_list)

        if uitype.UI_ROLE_SERVICE in roles:
            reply.extend(service_list)

        return reply

    @staticmethod
    def load_terminal(term_uiref: str) -> dict:
        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(uitype.UI_TYPE_TERMINAL)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        uiconf_pri = uiconf['ro']
        ext_uiconf_pub = uiconf['rw']
        ext_uiconf_pub['origin-guid'] = uiconf_pri['uuid']
        reply = {
            'type': uitype.UI_TYPE_TERMINAL,
            'conf': ext_uiconf_pub
        }
        return reply

    @staticmethod
    def save_terminal(term_uiref: str, new_uiconf: dict) -> dict:
        new_uiconf_conf = new_uiconf.get('conf')

        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(uitype.UI_TYPE_TERMINAL)

        uiconf = uimod.load_conf(default=None)
        if uiconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        fs.validate_json(new_uiconf_conf, uimod.schema_path)

        uiconf_pri: dict = uiconf['ro']
        uiconf_pub: dict = uiconf['rw']

        if uiconf_pri['uuid'] != new_uiconf_conf['origin-guid']:
            eventlog.log_d('fail: ParamsError (#/origin-guid is read-only)')
            raise ParamsError

        new_origin_name = new_uiconf_conf['origin-name']
        new_origin_place = new_uiconf_conf['origin-place']

        if uiconf_pub['origin-name'] != new_origin_name or uiconf_pub['origin-place'] != new_origin_place:
            term_uuid = TerminalUiconf(uimod.conf_path).get_uuid()
            ph.RootInfoUiconf.set_unit_name(uimod.ref, new_origin_name)

            from edge.support import db
            db.edit_term(term_uuid, new_origin_name, new_origin_place, manual=True)

            ph.TermAboutUiconf.save_conf(uimod.pid, term_uuid, term_place=new_origin_place, term_name=new_origin_name)

            uiconf_pub['origin-name'] = new_origin_name
            uiconf_pub['origin-place'] = new_origin_place

            uirefs = ph.RootInfoUiconf.get_linked_uirefs(uimod.ref)
            creator.generate(uimod.create_backup_dirpath(), uirefs, {uimod.ref: uiconf})

        reply = {
            'type': uitype.UI_TYPE_TERMINAL,
            'conf': new_uiconf_conf
        }
        return reply

    @classmethod
    def drop_terminal(cls, term_uiref: str) -> dict:
        root_info = ph.RootInfoUiconf.get_info()
        term_info = ph.InfoUiconf.find_unit_r(root_info, term_uiref)
        if term_info is None:
            eventlog.log_e('fail: ParamsError')
            raise ParamsError

        elif term_info.get('internal', True):
            eventlog.log_d('fail: ParamsError (terminal is internal)')
            raise ParamsError

        uimod = ph.UiMod.create(term_uiref)
        uimod.assert_type(uitype.UI_TYPE_TERMINAL)

        ext_terms_info = cls.drop_terminal2(uimod.pid)

        # restarting edgeserver with PCD
        eventlog.log_i('exec: restarting modules')
        dg.LifeCycle.sighup_ss_dgmod()
        dg.LifeCycle.restart_ws_dgmod()

        return ext_terms_info

    @classmethod
    def drop_terminal2(cls, term_pid: str) -> dict:
        from edge.webserver.sysconf.dgmod.es64.es64_module import es64_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.gc.gc_module import gc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.lr.lr_module import lr_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.mc.mc_module import mc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.mg.mg_module import mg_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.ms.ms_module import ms_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.msw.msw_module import msw_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.ra.ra_module import ra_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rc.rc_module import rc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rg.rg_module import rg_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rn.rn_module import rn_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.rs.rs_module import rs_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.sc.sc_module import sc_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.tl.tl_module import tl_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.tr.tr_module import tr_create_dgconf_deleter
        from edge.webserver.sysconf.dgmod.ws.ws_module import ws_create_dgconf_deleter

        dgconf_deleter_factory_fn_list: list[Callable[[IUiConfProvider], list[IDgConfDeleter]]] = [
            es64_create_dgconf_deleter,
            gc_create_dgconf_deleter,
            lr_create_dgconf_deleter,
            mc_create_dgconf_deleter,
            mg_create_dgconf_deleter,
            ms_create_dgconf_deleter,
            msw_create_dgconf_deleter,
            ra_create_dgconf_deleter,
            rc_create_dgconf_deleter,
            rg_create_dgconf_deleter,
            rn_create_dgconf_deleter,
            rs_create_dgconf_deleter,
            sc_create_dgconf_deleter,
            tl_create_dgconf_deleter,
            tr_create_dgconf_deleter,
            ws_create_dgconf_deleter
        ]

        factory_uimod = ph.UiMod(term_pid, uitype.UI_TYPE_FACTORY, '0')  # TODO
        entry_uiref = ph.UiMod.create_term(term_pid).ref
        root_info = ph.RootInfoUiconf.get_info()
        term_info = ph.InfoUiconf.find_unit_r(root_info, unit_ref=entry_uiref)
        if term_info is None:
            eventlog.log_i(f'fail find entry_uiref {entry_uiref}')
            raise NotImplementedError()

        entry_uirefs = ph.InfoUiconf.find_unit_refs_r(term_info)

        all_used_deleters = list[IDgConfDeleter]()
        for factory_fn in dgconf_deleter_factory_fn_list:
            dgconf_deleter = factory_fn(factory_uimod)
            all_used_deleters.extend(dgconf_deleter)

        for uiref in entry_uirefs:
            uimod = ph.RootInfoUiconf.get_uimods({uiref}, root_info).pop()  # TODO
            for factory_fn in dgconf_deleter_factory_fn_list:
                dgconf_deleter = factory_fn(uimod)
                all_used_deleters.extend(dgconf_deleter)

        list(map(lambda deleter: deleter.run(), all_used_deleters))

        # rm configs of ui-modules
        term_uimod = ph.UiMod.create_term(term_pid)

        shemas_dirpath = fs.join_path(ph.TERM_UISCHEMA_DIR, term_pid)
        ui_dirpath = fs.join_path(ph.TERM_UICONF_DIR, term_pid)
        list(map(fs.silent_rm_r, [shemas_dirpath, ui_dirpath]))

        terms_info, changed_uirefs = ph.RootInfoUiconf.del_unit_r(term_uimod.ref)
        ext_terms_info = ph.RootInfoUiconf.get_ext_root_info(terms_info)

        from edge.webserver.sysconf.dgconf.setupstore import SsDgconf
        SsDgconf.remove_ss_group(term_pid)

        reply = {
            'info': [ext_terms_info]
        }
        return reply

    @staticmethod
    def get_ra_type(term_pid: str) -> str | None:
        location_uimod = ph.UiMod(term_pid, uitype.UI_TYPE_LOCATION, '0')
        try:
            conf = location_uimod.load_conf()
            conf_ro = conf.get('ro') or dict()
            return conf_ro.get('type')

        except FsError:
            pass


def _main():
    # ph.RootInfoUiconf._CONF_PATH \
    #     = '/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/root.json'
    # factory_uiconf = fs.load_json3(
    #     '/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/mv_term.json')

    ph.RootInfoUiconf._CONF_PATH = (
        '/Users/d.feskov/CLionProjects/edgeserver/'
        '_share/edge-python-tools/edge-python-tools/share/misc/root.json')
    factory_uiconf: dict = fs.load_json3(
        '/Users/d.feskov/CLionProjects/edgeserver/'
        '_share/edge-python-tools/edge-python-tools/share/misc/mv_term.json')

    factory_info = factory_uiconf['pub']['info']
    factory_links = factory_uiconf['pri']['links']
    TerminalApi.create_term('controller.0', dict())


if __name__ == '__main__':
    _main()
