from typing import Final, List

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfBuilder, UirefToConfMap, Uideps, IDgConfDeleter
from edge.webserver.sysconf.dgmod.gc.gc_main import MainConfBuilder
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod import uitype


# noinspection PyUnusedLocal
def gc_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Uideps,
        uiref_to_newconf_map: UirefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {uitype.UI_TYPE_MV_CAMERA}):
        return []

    try:
        uimod_cam = target_uideps.find_any({uitype.UI_TYPE_MV_CAMERA})
        uimod_lpr = target_uideps.find_any({uitype.UI_TYPE_LPR})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref)
    ])

    if not activate:
        return []

    return [_GcCreator(uimod_cam, uimod_lpr)]


class _GcCreator(IDgConfBuilder):
    def __init__(self, uimod_cam: IUiConfProvider, uimod_lpr: IUiConfProvider) -> None:
        self._dgmod_gc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_GRABCONTROL, uimod_cam.id)

        self._uimod_cam: Final[IUiConfProvider] = uimod_cam
        self._uimod_lpr: Final[IUiConfProvider] = uimod_lpr

        self._main_active = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._dgmod_gc

    def setup(self, uiref_to_conf_map: UirefToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._uimod_cam.pid)

        cam_uiconf = uiref_to_conf_map.get(self._uimod_cam.ref) or self._uimod_cam.load_conf()
        cam_uiconf_control = cam_uiconf['rw']['videostream']['control']

        cam_uiconf_pri = cam_uiconf['ro']

        #
        exp_time_conf = [0, 0, 0]
        gain_conf = [0, 0, 0]
        gain_step = 1.0
        exptime_step = 20
        if cam_uiconf_control['profile'] == 'auto-70-150':
            gain_conf = [cam_uiconf_control['gain-auto-min'], cam_uiconf_control['gain-auto-max'], gain_step]
            exp_time_conf = [
                cam_uiconf_control['exposure-auto-min'],
                cam_uiconf_control['exposure-auto-max'],
                exptime_step]
        #

        gamma_limits = cam_uiconf_pri.get('sweep', dict()).get('gamma') or [0.4, 1.5, 0.1]
        main_builder = MainConfBuilder(gamma_limits)
        ir_jnode = cam_uiconf_pri.get('ir')
        pin_ref = None
        if ir_jnode is not None:
            pin_ref = ir_jnode['pin']

        control_ir = cam_uiconf_pri.get('control-ir', True)
        if pin_ref is not None and control_ir:
            main_builder.setup_ir_control(pin_ref)

        main_builder.setup_gain_conf(gain_conf)
        main_builder.setup_exp_time_conf(exp_time_conf)
        main_builder.setup_add_scene_img_react(spool_dir, f'mg{self._uimod_cam.id}-gc{self._uimod_cam.id}-scene')
        main_builder.setup_add_scene_aoi_react(spool_dir, f'lr{self._uimod_lpr.id}-gc{self._uimod_cam.id}-scene')
        main_builder.setup_add_scene_control_event(f'/etc/edge/mediagrabber.{self._uimod_cam.id}/dyn/')

        main_conf = main_builder.build()
        ss_active = cam_uiconf_control['profile'] == 'auto-70-150'

        self._main_active = main_conf, ss_active

    def build(self) -> None:
        main_conf, ss_active = self._main_active

        dgmod = self._dgmod_gc
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _GcDeleter(IDgConfDeleter):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_gc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_GRABCONTROL, uimod_cam.id)

    def run(self) -> None:
        self._dgmod_gc.uninstall()

def gc_create_dgconf_deleter(target_uimod: IUiConfProvider) -> List[IDgConfDeleter]:
    if target_uimod.type == uitype.UI_TYPE_MV_CAMERA:
        return [_GcDeleter(target_uimod)]
    return []
