from typing import Final, List

from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter,IDgConfBuilder, UirefToConfMap, Uideps
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uimod import uitype


# noinspection PyUnusedLocal
def ws_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Uideps,
        uiref_to_newconf_map: UirefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {uitype.UI_TYPE_WEBSERVER}):
        return []

    uimods_term = target_uideps.try_find_all(uitype.UI_TYPE_TERMINAL)
    uimods_hstreaming = target_uideps.try_find_all(uitype.UI_TYPE_HTTP_STREAMING)

    assert len(uimods_hstreaming) == len(uimods_term)

    uimods_term.sort(key=lambda item: int(item.id))
    uimods_hstreaming.sort(key=lambda item: int(item.id))

    activate = forced or any([
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_hstreaming])
    ])

    if not activate:
        return []

    return [_WsTermCreator(uimods_term, uimods_hstreaming)]


class _WsTermCreator(IDgConfBuilder):
    def __init__(self, terminal_uimods: list[IUiConfProvider], streaming_uimods: list[IUiConfProvider]) -> None:
        self._dgmod_ws: Final[dg.DgMod] = dg.CTRL_WS_DGMOD

        self._uimods_hstreaming: Final[list[IUiConfProvider]] = streaming_uimods
        self._uimods_terminal: Final[list[IUiConfProvider]] = terminal_uimods

        self._main_conf = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._dgmod_ws

    def setup(self, uiref_to_newconf_map: UirefToConfMap) -> None:
        main_conf = self._dgmod_ws.load_main_conf()

        media_jnode: list[dict] = []
        for index in range(0, len(self._uimods_terminal)):
            streaming_uimod = self._uimods_hstreaming[index]
            terminal_uimod = self._uimods_terminal[index]

            hstreaming_uiconf = uiref_to_newconf_map.get(streaming_uimod.ref) or streaming_uimod.load_conf()

            terminal_uiconf = TerminalUiconf(terminal_uimod.conf_path)
            terminal_uuid = terminal_uiconf.get_uuid()
            origin_jnode = {
                'origin': terminal_uuid
            }

            if hstreaming_uiconf['rw']['enabled']:
                spool_dir = ph.get_spool_dir(terminal_uimod.pid)
                origin_jnode['LiveEndpoint'] = fs.join_path(spool_dir, f'or{streaming_uimod.id}-ws0-scene')

            media_jnode.append(origin_jnode)

        main_conf['Media'] = media_jnode
        self._main_conf = main_conf

    def build(self) -> None:
        main_conf: dict = self._main_conf

        dgmod = self._dgmod_ws
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        dg.LifeCycle.restart_ws_dgmod()


class _WsTermDeleter(IDgConfDeleter):
    def __init__(self, term_uuid: str) -> None:
        self._dgmod_ws: Final[dg.DgMod] = dg.CTRL_WS_DGMOD
        self._term_uuid = term_uuid

    def run(self) -> None:
        dgconf = self._dgmod_ws.load_main_conf()
        dgconf['Media'] = [x for x in dgconf['Media'] if x['origin'] != self._term_uuid]
        self._dgmod_ws.save_main_conf(dgconf)


def ws_create_dgconf_deleter(target_uimod: IUiConfProvider) -> List[IDgConfDeleter]:
    if target_uimod.type == uitype.UI_TYPE_FACTORY:
        term_uimod = ph.UiMod.create_term(target_uimod.pid)
        term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
        return [_WsTermDeleter(term_uuid)]

    return []
