import os
import errno

from collections import defaultdict
from collections.abc import Iterable

import edge.webserver.sysconf.uimod.uitype
from edge import eventlog
from edge.types.exceptions import InternalError
from edge.webserver.sysconf import ph, file_lock
from edge.webserver.sysconf.backup import IBackup, DgmodBackup, UimodBackup, NewFileBackup
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfBuilder, UirefToConfMap, UirefToUimodMap, Uideps
from edge.webserver.sysconf.types.conf import IDgConfProvider, IUiConfProvider


def _locked_generate(backup_dir: str,
                     dgconf_builders: list[IDgConfBuilder],
                     used_uimods: set[IUiConfProvider],
                     used_dgmods: set[IDgConfProvider],
                     uiref_to_newconf_map: UirefToConfMap) -> None:
    try:
        os.makedirs(backup_dir)
    except os.error as x:
        if x.errno != errno.EEXIST:
            eventlog.log_e(f'fail: makedirs {backup_dir} ({errno.errorcode[x.errno]})')
        raise NotImplementedError()

    backup_error = False

    dgmods_backup = list[IBackup]()
    for dgmod in used_dgmods:
        if dgmod.ref == dg.CTRL_SS_DGMOD.ref:
            continue

        dgmod_backup = DgmodBackup.create(backup_dir, dgmod)
        if dgmod_backup is None:
            eventlog.log_w(f'fail: backup dgmod (ref:{dgmod.ref})')
            backup_error = True
            continue

        dgmods_backup.append(dgmod_backup)

    uimods_backup = list[IBackup]()
    for uimod in used_uimods:
        if uimod.type in edge.webserver.sysconf.uimod.uitype.FILELESS_UI_TYPES:
            continue

        uimod_backup = UimodBackup.create(backup_dir, uimod)
        if uimod_backup is None:
            eventlog.log_w(f'fail: backup uimod (ref:{uimod.ref})')
            backup_error = True
            continue

        uimods_backup.append(uimod_backup)

    for uiref, newconf in uiref_to_newconf_map.items():
        uimod = list(filter(lambda x: x.ref == uiref, used_uimods))[0]
        if uimod.type in edge.webserver.sysconf.uimod.uitype.FILELESS_UI_TYPES:
            continue

        uimod_backup = NewFileBackup.create(backup_dir, uimod, newconf)
        if uimod_backup is None:
            eventlog.log_w(f'fail: backup newconf (ref:{uimod.ref})')
            backup_error = True

        uimods_backup.append(uimod_backup)

    if backup_error:
        raise InternalError()

    # noinspection PyBroadException
    try:
        restore_backup = False
        try:
            list(map(lambda builder: builder.setup(uiref_to_newconf_map), dgconf_builders))
        except Exception:
            eventlog.log_d('fail')
            raise

        try:
            list(map(lambda builder: builder.build(), dgconf_builders))
        except Exception:
            eventlog.log_d('fail')
            raise

        for uiref, newconf in uiref_to_newconf_map.items():
            uimod = list(filter(lambda x: x.ref == uiref, used_uimods))[0]
            uimod.save_conf(newconf)

        # noinspection PyBroadException
        try:
            for dgmod_backup in dgmods_backup:
                dgmod_backup.remove()

            for uimod_backup in uimods_backup:
                uimod_backup.remove()

            if not os.listdir(backup_dir):
                os.rmdir(backup_dir)

        except Exception:
            eventlog.log_x()

    except Exception:
        restore_backup = True
        eventlog.log_x()

    if restore_backup:
        # TODO: what if except inside?
        list(map(lambda bak: bak.restore(), uimods_backup))
        list(map(lambda bak: bak.restore(), dgmods_backup))
        raise InternalError()


def generate(backup_dir: str,
             affected_uirefs: Iterable[str],
             uiref_to_newconf_map: UirefToConfMap | None = None,
             forced: bool = False
             ) -> set[dg.DgMod]:
    """
    :param backup_dir:
        Absolute path to the backup directory.
    :param affected_uirefs:
        A list of units that may be affected by the changes.
            - The created unit and its direct dependencies (linked units).
            - The modified unit and its direct dependencies (linked units).
            - The direct dependencies (linked units) of the deleted unit.
    :param uiref_to_newconf_map:
    :param forced:
    """
    uiref_to_newconf_map = uiref_to_newconf_map or dict()

    from edge.webserver.sysconf.dgmod.bs64.bs64_module import bs64_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.es64.es64_module import es64_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.gc.gc_module import gc_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.lr.lr_module import lr_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.mc.mc_module import mc_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.mg.mg_module import mg_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.ms.ms_module import ms_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.msw.msw_module import msw_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.ra.ra_module import ra_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.rc.rc_module import rc_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.rg.rg_module import rg_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.rn.rn_module import rn_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.rs.rs_module import rs_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.sc.sc_module import sc_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.tl.tl_module import tl_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.tr.tr_module import tr_create_dgconf_builder
    from edge.webserver.sysconf.dgmod.ws.ws_module import ws_create_dgconf_builder

    dgconf_builder_factory_fn_list = [
        bs64_create_dgconf_builder,
        es64_create_dgconf_builder,
        gc_create_dgconf_builder,
        lr_create_dgconf_builder,
        mc_create_dgconf_builder,
        mg_create_dgconf_builder,
        ms_create_dgconf_builder,
        msw_create_dgconf_builder,
        ra_create_dgconf_builder,
        rc_create_dgconf_builder,
        rg_create_dgconf_builder,
        rn_create_dgconf_builder,
        rs_create_dgconf_builder,
        sc_create_dgconf_builder,
        tl_create_dgconf_builder,
        tr_create_dgconf_builder,
        ws_create_dgconf_builder
    ]

    root_info, root_links = ph.RootInfoUiconf.get_info_links()

    all_used_uimods = set[IUiConfProvider]()
    all_used_builders = list[IDgConfBuilder]()

    for affected_uiref in affected_uirefs:
        # define the direct dependencies (linked units) of the affected unit.
        linked_uirefs = ph.RootInfoUiconf.get_linked_uirefs(affected_uiref, root_links)
        if not linked_uirefs:
            linked_uirefs = [affected_uiref]

        linked_uimods = ph.RootInfoUiconf.get_uimods(linked_uirefs, root_info)

        # group the direct dependencies (linked units) of the affected unit by type.
        type_to_uiref_to_uimod_map = defaultdict[str, UirefToUimodMap](UirefToUimodMap)
        for lindked_uimod in linked_uimods:
            type_to_uiref_to_uimod_map[lindked_uimod.type][lindked_uimod.ref] = lindked_uimod

        uideps = Uideps(type_to_uiref_to_uimod_map)

        for factory_fn in dgconf_builder_factory_fn_list:
            dgconf_builder = factory_fn(affected_uiref, uideps, uiref_to_newconf_map, forced)
            all_used_builders.extend(dgconf_builder)

        all_used_uimods.update(linked_uimods)

    all_used_dgmods = set([x.dgmod for x in all_used_builders])
    answer = all_used_dgmods
    if forced:
        all_used_dgmods = {dg.CTRL_ES_DGMOD, dg.CTRL_SS_DGMOD, dg.CTRL_WS_DGMOD}

    lock = file_lock.UimodsLocker(all_used_uimods)
    try:
        _locked_generate(backup_dir, all_used_builders, all_used_uimods, all_used_dgmods, uiref_to_newconf_map)
    finally:
        lock.release()

    list(map(lambda x: x.notify(), all_used_builders))

    return answer
