import os

from collections.abc import Callable, Iterable
from datetime import datetime
from functools import reduce
from typing import Any, Final, Self

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.types.mmr import MMR_CLASSES
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uiconf.uiconf2 import UiConf2
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uiconf.rtsp_streaming import RtspStreamingUiconf
from edge.webserver.sysconf.uimod.eventstore import EsUimod
from edge.webserver.sysconf.uimod import uitype

SEMVER_PATH = '/opt/edge/semver'
CTRL_SPOOL_DIR = '/var/spool/edge/'
CTRL_ARCH_DIR = '/var/opt/edge/'
CTRL_BACKUP_DIR = '/etc/edge/setupstore/backup/'
CTRL_DGCONF_DIR = '/etc/edge/'
CTRL_UICONF_DIR = '/etc/edge/setupstore/admin/actions/'
CTRL_UICONF_PATH = '/etc/edge/setupstore/about.json'
CTRL_UISCHEMA_DIR = '/etc/edge/setupstore/schemas/'
CTRL_DEF_UICONF_DIR = '/opt/edge/share/setupstore/ui/'
CTRL_DEF_DGCONF_DIR = '/opt/edge/share/etc/edge/'

TERM_UICONF_DIR = fs.join_path(CTRL_UICONF_DIR, 'pipeline')
TERM_UISCHEMA_DIR = fs.join_path(CTRL_UISCHEMA_DIR, 'pipeline')
TERM_DEF_UICONF_DIR = fs.join_path(CTRL_DEF_UICONF_DIR, 'pipeline')


class UiMod(IUiConfProvider):
    def __init__(self, uimod_pid: str, uimod_type: str, uimod_id: str) -> None:
        # self._ref: Final[str] = '{}.{}'.format(self.type, str(10 * int(self.pid) + int(self.id)))
        self._ref: Final[str] = f'{uimod_type}.{uimod_pid}_{uimod_id}'

        self._type: Final[str] = uimod_type
        self._pid: Final[str] = uimod_pid
        self._id: Final[str] = uimod_id

        self._conf = None

    def __hash__(self) -> int:
        return hash(self.type + self.id)

    def __str__(self) -> str:
        return self.ref

    def __eq__(self, other: Self) -> bool:
        return (isinstance(other, self.__class__) and
                getattr(other, 'pid', None) == self.pid and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'id', None) == self.id)

    @classmethod
    def create(cls, uimod_ref: str) -> Self:
        try:
            uimod_type, uimod_pid_id = uimod_ref.split('.', 1)
            uimod_pid, uimod_id = uimod_pid_id.split('_', 1)
        except Exception:
            eventlog.log_d(f'fail: NotFoundError ({uimod_ref})')
            raise NotFoundError

        return cls(uimod_pid, uimod_type, uimod_id)

    @classmethod
    def create_term(cls, term_pid: str) -> Self:
        return cls(term_pid, uitype.UI_TYPE_TERMINAL, str(int(term_pid) * 10))

    @classmethod
    def create_controller(cls, controller_id: str = '0') -> Self:
        return cls(controller_id, uitype.UI_TYPE_CONTROLLER, str(int(controller_id) * 10))

    def assert_type(self, uimod_type: str) -> None:
        if self.type != uimod_type:
            eventlog.log_d(f'fail: NotFoundError ({self.ref})')
            raise NotFoundError

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def type(self) -> str:
        return self._type

    @property
    def pid(self) -> str:
        return self._pid

    @property
    def id(self) -> str:
        return self._id

    @property
    def conf_dir(self) -> str:
        return fs.join_path(TERM_UICONF_DIR, self.pid, self.type)

    @property
    def conf_path(self) -> str:
        old_pid, old_id = divmod(int(self.id), 10)
        return fs.join_path(self.conf_dir, f'{old_id}.json')
        # return fs.join_path(self.conf_dir, '{}.json'.format(self.id))

    def load_conf(self, **kwargs) -> dict:
        """
        :raise FsError:
        """
        if self._conf is None:
            self._conf = fs.load_json3(self.conf_path, **kwargs)

        return self._conf

    def save_conf(self, conf_jnode: dict) -> None:
        fs.write_json_file(conf_jnode, self.conf_path, pretty=True)

    @property
    def schema_path(self) -> str:
        old_pid, old_id = divmod(int(self.id), 10)
        return fs.join_path(TERM_UISCHEMA_DIR, self.pid, self.type, f'{old_id}.json')
        # return fs.join_path(BASE_UI_PIPELINE_SCHEMA_DIRPATH, self.pid, self.type, '{}.json'.format(self.id))

    @property
    def schema_limits(self) -> dict | None:
        schema_jnode = fs.load_json3(self.schema_path)
        return schema_jnode.get('xLimits')

    @property
    def lock_path(self) -> str:
        return fs.join_path('/var/run/edge', '{}.lock').format(self.ref)

    def create_backup_dirpath(self) -> str:
        import time
        backup_action = '{:d}-{}'.format(int(round(time.time() * 1e6)), self.ref)
        backup_dir = fs.join_path(CTRL_BACKUP_DIR, backup_action)
        return backup_dir

    def get_defconf_path(self, factory_conf: dict | None = None) -> str:
        import fnmatch

        if self.type == uitype.UI_TYPE_MV_CAMERA:
            if factory_conf is None:
                eventlog.log_e_param('factory_conf')
                raise NotImplementedError

            cam_vendor = factory_conf['vendor'].lower()  # TODO workaround for migrate script
            cam_model = str.replace(str(factory_conf['model']), ' ', '-')

            model_files = os.listdir(fs.join_path(TERM_DEF_UICONF_DIR, uitype.UI_TYPE_MV_CAMERA))
            model_files2 = fnmatch.filter(model_files, f'{cam_vendor}_{cam_model}*jnode.json')
            for model_file in model_files2:
                # TODO: what to do if there are jnodes for one camera with different focal-lengths?
                return fs.join_path(TERM_DEF_UICONF_DIR, self.type, model_file)

            eventlog.log_e(f'fail: unsupported (vendor:{cam_vendor}, model:{cam_model})')
            raise NotImplementedError

        #
        type_to_name_map: dict[str, str] = dict()
        if self.type != uitype.UI_TYPE_TERMINAL:
            root_info = RootInfoUiconf.get_info()
            _, term_info = InfoUiconf.find_unit_parent_r(root_info, unit_ref=self.ref, parent_type=uitype.UI_TYPE_TERMINAL)
            if term_info is None:
                eventlog.log_e(f'fail: find_unit_parent_r (ref:{self.ref}, parent_type:{uitype.UI_TYPE_TERMINAL})')
                raise NotImplementedError

            elif InfoUiconf.find_unit_r(term_info, unit_type=uitype.UI_TYPE_RADAR):
                type_to_name_map = {
                    uitype.UI_TYPE_HTTP_STREAMING: 'speed-jnode.json',
                    uitype.UI_TYPE_RTSP_STREAMING: 'speed-jnode.json',
                    uitype.UI_TYPE_LOCATION: 'ummr0a-type29-jnode.json',
                    uitype.UI_TYPE_DET_TRANSIT: 'def-speed-jnode.json',
                    uitype.UI_TYPE_RTSP_CAMERA: 'rtsp-jnode.json'
                }

            elif InfoUiconf.find_unit_r(term_info, unit_type=uitype.UI_TYPE_MV_CAMERA):
                type_to_name_map = {
                    uitype.UI_TYPE_HTTP_STREAMING: 'lpr-jnode.json',
                    uitype.UI_TYPE_RTSP_STREAMING: 'lpr-jnode.json'
                }

            elif InfoUiconf.find_unit_r(term_info, unit_type=uitype.UI_TYPE_IPCOM_WAGA):
                type_to_name_map = {
                    uitype.UI_TYPE_HTTP_STREAMING: 'lpr-jnode.json',
                    uitype.UI_TYPE_RTSP_CAMERA: 'rtsp-jnode.json',
                    uitype.UI_TYPE_DET_TRANSIT: 'def-speed-jnode.json'
                }

            elif InfoUiconf.find_unit_r(term_info, unit_type=uitype.UI_TYPE_RTSP_CAMERA):
                type_to_name_map = {
                    uitype.UI_TYPE_HTTP_STREAMING: 'lpr-jnode.json',
                    uitype.UI_TYPE_RTSP_CAMERA: 'rtsp-jnode.json'
                }

            else:
                eventlog.log_e('unsupported')
                raise NotImplementedError

        conf_name = type_to_name_map.get(self.type, 'def-jnode.json')
        return fs.join_path(TERM_DEF_UICONF_DIR, self.type, conf_name)

    def _install_conf(self, factory_conf: dict | None = None) -> None:
        eventlog.log_i(f'_install_conf {self.ref}')
        conf_path = self.conf_path
        if fs.path_exists(conf_path):
            eventlog.log_w('fail: already installed')
            return

        defconf_path = self.get_defconf_path(factory_conf)

        fs.silent_mkdir_r(self.conf_dir)
        fs.copy_file(defconf_path, conf_path)

        if self.type == uitype.UI_TYPE_MV_CAMERA:
            uiconf = self.load_conf()
            uiconf['rw']['serial'] = str(factory_conf['serial'])
            self.save_conf(uiconf)

        elif self.type == uitype.UI_TYPE_RTSP_CAMERA:
            if factory_conf.get('with-intrinsic', False):
                uiconf = self.load_conf()
                uiconf['rw']['conf']['intrinsic'] = {
                    'active': False,
                    'frustum': {
                        'pixel-size': 3.45,
                        'focal-dist': 50,
                        'sensor-size': [
                            1920, 1200
                        ]
                    }
                }
                self.save_conf(uiconf)

        elif self.type == uitype.UI_TYPE_TERMINAL:
            uiconf = self.load_conf()
            uiconf['ro']['uuid'] = factory_conf['term-uuid']
            uiconf['ro']['ctime'] = datetime.now().isoformat()
            uiconf['rw']['origin-name'] = self.ref
            uiconf['rw']['origin-place'] = factory_conf['term-pid']
            self.save_conf(uiconf)
            RootInfoUiconf.set_unit_name(self.ref, self.ref)

        elif self.type == uitype.UI_TYPE_RTSP_STREAMING:
            uiconf = self.load_conf()
            uiconf['rw']['streaming']['tcp-port'] += int(factory_conf['term-pid'])
            uiconf['rw']['streaming']['url'] = RtspStreamingUiconf(uiconf['rw']).rtsp_url
            self.save_conf(uiconf)

        eventlog.log_i(f'done: cp {defconf_path} {conf_path}')

    def install_schema(self) -> None:
        conf_path: Final[str] = self.conf_path
        if fs.path_exists(conf_path):
            fs.ln_schema(self.schema_path, conf_path)

    def install(self, factory_conf: dict | None = None) -> None:
        if self.type not in uitype.FILELESS_UI_TYPES:
            self._install_conf(factory_conf)
            self.install_schema()

    def uninstall(self) -> None:
        if self.type not in uitype.FILELESS_UI_TYPES:
            fs.silent_rm_r(self.conf_path)
            fs.silent_rm_r(self.schema_path)


class UiMod2(IUiConfProvider):
    def __init__(self, ref: str) -> None:
        assert ref

        self._ref: Final[str] = ref
        self._conf_jnode: dict | None = None
        self._container_path: str | None = None

    def __eq__(self, other: Self) -> bool:
        return isinstance(other, self.__class__) and getattr(other, 'ref', None) == self.ref

    def __hash__(self) -> int:
        return hash(self.ref)

    def __str__(self) -> str:
        return self.ref

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def type(self) -> str:
        segments = self.ref.split('~')
        subtypes = [x.split('.')[0] for x in segments]
        return '~'.join(subtypes)

    @property
    def pid(self) -> str:
        eventlog.log_e('fail: unsupported')
        raise NotImplementedError()

    @property
    def id(self) -> str:
        eventlog.log_e('fail: unsupported')
        raise NotImplementedError()

    @property
    def lock_path(self) -> str:
        return self.uiconf.lock_path

    def create_backup_dirpath(self) -> str:
        import time
        backup_action = '{:d}-{}'.format(int(round(time.time() * 1e6)), self.ref)
        backup_dir = fs.join_path(CTRL_BACKUP_DIR, backup_action)
        return backup_dir

    def assert_ref(self) -> None:
        """
        :raises NotFoundError:
        """
        self.load_conf()

    def assert_type(self, other_type: str) -> None:
        """
        :raises NotFoundError:
        """
        if self.type != other_type:
            eventlog.log_d(f'fail: NotFoundError ({self.ref})')
            raise NotFoundError

    @property
    def root_ref(self) -> str:
        """
        :raises NotFoundError:
        """
        return self.ref.split('~', 1)[0]

    @property
    def parent_ref(self) -> str:
        """
        :raises NotFoundError:
        """
        parent_ref = '~'.join(self.ref.split('~')[:-1])
        if not parent_ref:
            raise NotFoundError

        return parent_ref

    @property
    def parent(self) -> Self:
        """
        :raises NotFoundError:
        """
        return UiMod2(self.parent_ref)

    def child_ref(self, child_type: str, child_chan: str | int) -> str:
        if not child_type.startswith(self.type):
            raise NotFoundError

        child_type_segments = child_type.split('~')
        if not len(self.type.split('~')) == len(child_type_segments) - 1:
            raise NotFoundError

        return f'{self.ref}~{child_type_segments[-1]}.{child_chan}'

    def child(self, child_type: str, child_chan: str | int) -> Self:
        return UiMod2(self.child_ref(child_type, child_chan))

    def load_conf(self) -> dict:
        """
        :raise NotFoundError:
        """
        if None in [self._container_path, self._conf_jnode]:
            container_names_desc = []
            segments = self.ref.split('~')
            for index in range(0, len(segments)):
                container_name = '~'.join(segments[0:index + 1]) + '.json'
                container_names_desc.append(container_name)

            for container_name in reversed(container_names_desc):
                container_path = fs.join_path(CTRL_UICONF_DIR, container_name)
                container_jnode = fs.load_json3(container_path, default=None) or dict()
                conf_jnode = container_jnode.get(self.ref)
                if conf_jnode:
                    self._container_path = container_path
                    self._conf_jnode = conf_jnode
                    return conf_jnode

            raise NotFoundError

        return self._conf_jnode

    @property
    def schema_path(self) -> str:
        eventlog.log_e('fail: unsupported')
        raise NotImplementedError()

    @property
    def conf_path(self) -> str:
        import warnings
        warnings.warn('message', PendingDeprecationWarning, stacklevel=2)

        self.assert_ref()
        return self._container_path

    def save_conf(self, conf_jnode: dict) -> None:
        self.assert_ref()
        container_jnode = fs.load_json3(self._container_path, default=None) or dict()
        container_jnode[self.ref] = conf_jnode
        fs.write_json_file(container_jnode, self._container_path, pretty=True)

        self._conf_jnode = conf_jnode

    @property
    def uiconf(self) -> UiConf2:
        """
        :raise NotFoundError:
        """
        return UiConf2(self.load_conf())

    def save_uiconf(self, conf: UiConf2) -> None:
        """
        :rtype: None
        """
        with open(SEMVER_PATH, 'r') as f:
            ver = f.readlines()[0].strip()

        conf.set_ver(ver)
        self.save_conf(conf.raw)

    def install(self, conf: UiConf2, container_ref: str | None = None) -> None:
        try:
            self.assert_ref()
            eventlog.log_e('fail: InternalError (already installed)')
            raise InternalError

        except NotFoundError:
            pass

        if not self.ref.startswith(container_ref):
            raise NotImplementedError

        container_name = f'{container_ref or self.ref}.json'
        container_path = fs.join_path(CTRL_UICONF_DIR, container_name)
        container_jnode = fs.load_json3(container_path, default=None) or dict()
        container_jnode[self.ref] = conf.raw
        fs.write_json_file(container_jnode, container_path, pretty=True)

        self._container_path = container_path
        self._conf_jnode = conf.raw

    def uninstall(self) -> None:
        self.assert_ref()
        container_jnode = fs.load_json3(self._container_path, default=None) or dict()
        del container_jnode[self.ref]
        if len(container_jnode) > 0:
            fs.write_json_file(container_jnode, self._container_path, pretty=True)
        else:
            fs.silent_rm_r(self._container_path)


def get_spool_dir(ui_pid: str) -> str:
    return fs.join_path(CTRL_SPOOL_DIR, f'pipeline-{ui_pid}')


# noinspection PyClassHasNoInit
class TermAboutUiconf:
    @staticmethod
    def get_conf_path(ui_pid: str) -> str:
        return fs.join_path(TERM_UICONF_DIR, ui_pid, 'about.json')

    @classmethod
    def save_conf(cls, term_pid: str, term_uuid: str,
                  term_ctime: str | None = None,
                  term_place: str | None = None,
                  term_name: str | None = None) -> None:

        conf_path = cls.get_conf_path(term_pid)
        conf = fs.load_json3(conf_path, default={
            'origin': dict()
        })

        conf_origin: dict = conf['origin']
        conf_origin['uuid'] = term_uuid
        conf_origin['place'] = term_place or term_pid
        conf_origin['name'] = term_name or term_pid

        if term_ctime:
            conf_origin['time'] = term_ctime

        elif 'time' not in conf_origin:
            conf_origin['time'] = datetime.now().isoformat()

        fs.write_json_file(conf, conf_path)


# noinspection PyClassHasNoInit
class InfoUiconf:
    @classmethod
    def get_info_links(cls, info_path: str) -> tuple[dict, list[tuple[str, str]]]:
        uiconf = fs.load_json3(info_path, default=None)
        if uiconf is None:
            raise NotImplementedError

        uiconf_pri: dict = uiconf['pri']
        uiconf_pub: dict = uiconf['pub']

        root_links: list[tuple[str, str]] = uiconf_pri['links']
        root_info: dict = uiconf_pub['info']

        return root_info, root_links

    @classmethod
    def map_units_r(cls, unit_info: dict,
                    map_fn: Callable[[dict], Any] | None = None,
                    unit_types: list[str] | None = None) -> list[Any]:
        """
        Finds and applies a function to all units which satisfy the constraints.

        The root unit is included in the result if it satisfies the constraints.

        :param unit_info:
            The root unit info.
        :param map_fn:
            The function to be applied to each matched unit.
            The default function returns the unit info.
        :param unit_types:
            Unit type constraint. All units are matched if not specified.
        """

        def find_units_r(unit: dict, result: list[dict]) -> list[dict]:
            subunits = unit.get('units', [])
            for subunit in subunits:
                find_units_r(subunit, result)

            if unit_types is None:
                result.append(unit)

            elif unit['type'] in unit_types:
                result.append(unit)

            return result

        subunits_info = find_units_r(unit_info, [])
        return list(map(map_fn, subunits_info)) if map_fn else subunits_info

    @classmethod
    def find_unit_refs_r(cls, unit_info: dict, unit_types: list[str] | None = None) -> list[str]:
        """
        Finds all unit refs which satisfy the constraints.

        The root unit is included in the result if it satisfies the constraints.

        :param unit_info:
            The root unit info.
        :param unit_types:
            Unit type constraint. All units are matched if not specified.
        """
        return cls.map_units_r(unit_info, lambda x: x['ref'], unit_types)

    @classmethod
    def find_unit_r(cls, root_info: dict,
                    unit_ref: str | None = None,
                    unit_type: str | None = None) -> dict | None:

        assert any([unit_ref, unit_type])

        if root_info['ref'] == unit_ref or root_info['type'] == unit_type:
            return root_info

        subunits = root_info.get('units', [])
        for subunit in subunits:
            unit = cls.find_unit_r(subunit, unit_ref, unit_type)
            if unit:
                return unit

    @classmethod
    def find_unit_parent_r(cls, root_info: dict,
                           unit_ref: str | None = None,
                           unit_type: str | None = None,
                           parent_type: str | None = None
                           ) -> tuple[dict, dict] | tuple[dict, dict | None] | tuple[None, None]:
        """
        Finds a unit and one of its parents in the root_info tree.

        At least one of the arguments must be specified: unit-type, unit-ref

        :param root_info:
            The root unit info
        :param unit_ref:
            Unit ref constraint.
        :param unit_type:
            Unit type constraint.
        :param parent_type:
            Parent unit type constraint.
            If not provided, the most nested parent will be returned.
        """
        assert any([unit_ref, unit_type])

        def find_unit_parents_r(unit: dict, parents: list[dict]) -> tuple[dict | None, list[dict]]:
            """
            Finds a unit and all its parents.
            :param unit:
                Current unit info
            :param parents:
                List of the parent units.
                The order of parents is from the most nested to the least nested.
            """
            if unit['ref'] == unit_ref or unit['type'] == unit_type:
                return unit, parents

            subunits = unit.get('units', [])
            for subunit in subunits:
                found_unit, _ = find_unit_parents_r(subunit, parents)
                if found_unit:
                    parents.append(unit)
                    return found_unit, parents

            return None, parents

        unit_info, parents_info = find_unit_parents_r(root_info, [])
        for parent in parents_info:
            if parent_type is None:
                return unit_info, parent

            elif parent_type == parent['type']:
                return unit_info, parent

        return unit_info, None


# noinspection PyClassHasNoInit
class RootInfoUiconf(InfoUiconf):
    _DEFCONF_PATH = '/opt/edge/share/setupstore/ui/root-info.json'
    _CONF_PATH = '/etc/edge/setupstore/root-info.json'

    @classmethod
    def get_ver(cls) -> str | None:
        uiconf = fs.load_json3(cls._CONF_PATH, default=None)
        if uiconf is None:
            raise NotImplementedError

        return uiconf.get('ver')

    @classmethod
    def set_info_links(cls, root_info: dict,
                       root_links: list[tuple[str, str]],
                       ver: str | None = None) -> None:
        new_ver = ver or cls.get_ver()
        uiconf = {
            'ver': new_ver,
            'pri': {
                'links': root_links
            },
            'pub': {
                'info': root_info
            }
        }

        fs.write_json_file(uiconf, cls._CONF_PATH, True)

    @classmethod
    def get_info_links(cls, info_path: str | None = None) -> tuple[dict, list[tuple[str, str]]]:
        return InfoUiconf.get_info_links(info_path or cls._CONF_PATH)

    @classmethod
    def get_info(cls) -> dict:
        return cls.get_info_links()[0]

    @classmethod
    def get_links(cls) -> list[tuple[str, str]]:
        return cls.get_info_links()[1]

    @classmethod
    def set_unit_name(cls, unit_ref: str, unit_name: str) -> None:
        root_info, root_links = cls.get_info_links()
        unit_info = cls.find_unit_r(root_info, unit_ref)
        if unit_info is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_info['name'] = unit_name
        cls.set_info_links(root_info, root_links)

    @classmethod
    def get_linked_uirefs(cls, uiref: str, links: list[tuple[str, str]] | None = None) -> set[str]:
        """
        :param uiref:
            The endpoint that must to be a part of the matched link.
        :param links:
            The pre-loaded (cached) list of links. The list will be loaded if not provided.
        :return:
            A set of the unique endpoints of the matched links.
            The uiref endpoint is included in the result if at least one link matches.
        """
        if links is None:
            links = cls.get_links()

        used_uirefs = set[str]()
        for link in links:
            if uiref in link:
                used_uirefs.update(link)

        return used_uirefs

    @classmethod
    def get_uimods(cls, uirefs: Iterable[str], root_info: dict | None = None) -> set[IUiConfProvider]:
        if root_info is None:
            root_info = cls.get_info()

        uimods = set()
        for uiref in uirefs:
            uimod_info = cls.find_unit_r(root_info, unit_ref=uiref)
            if uimod_info is None:
                eventlog.log_w(f'fail: find_unit_r (uiref:{uiref})')
                continue

            unit_type = uimod_info.get('type')
            if unit_type == uitype.UI_TYPE_WEBSERVER:
                uimod = UiMod('0', uitype.UI_TYPE_WEBSERVER, '0')
            elif unit_type == uitype.UI_TYPE_EVENTSTORE:
                uimod = EsUimod.create(uimod_info['ref'])
            elif unit_type in uitype.UIMOD2_TYPES:
                uimod = UiMod2(uimod_info['ref'])
            else:
                uimod = UiMod.create(uimod_info['ref'])

            uimods.add(uimod)

        return uimods

    @classmethod
    def new_uimod_pid_type_id(cls, root_info: dict | None, uimod_type: str,
                              uimod_pid: str | None = None) -> tuple[str | None, str, str]:

        used_refs = cls.find_unit_refs_r(root_info, unit_types=[uimod_type])

        if uimod_type == uitype.UI_TYPE_TERMINAL:
            if uimod_pid is None:
                # 'terminal.1_10' => pid:1, id:10
                import sys
                used_pids = [int(UiMod.create(x).pid) for x in used_refs]
                for candidate in range(sys.maxsize):
                    if candidate not in used_pids:
                        uimod_pid = candidate
                        break

            uimod_id = 10 * int(uimod_pid)
            eventlog.log_d(f'done: pid:{uimod_pid}, type:{uimod_type}, id:{uimod_id}')
            return str(uimod_pid), uimod_type, str(uimod_id)

        elif root_info['type'] in [uitype.UI_TYPE_CONTROLLER, uitype.UI_TYPE_WIEGAND_PCB, uitype.UI_TYPE_IIT_CA]:
            # 'eventstore.14' => pid:None, id:14
            used_ids = [int(x.split('.')[-1]) for x in used_refs]
            for candidate in range(10):
                if candidate not in used_ids:
                    uimod_id = candidate
                    eventlog.log_d(f'done: pid:{None}, type:{uimod_type}, id:{uimod_id}')
                    return None, uimod_type, str(uimod_id)

        elif root_info['type'] in [uitype.UI_TYPE_TERMINAL, uitype.UI_TYPE_TRAFFIC_RULES,
                                   uitype.UI_TYPE_MV_CAMERA, uitype.UI_TYPE_RTSP_CAMERA, uitype.UI_TYPE_LPR]:
            # 'wrong-way.1_14' => 14 => 4
            used_uimod_rel_ids = cls.map_units_r(
                root_info, lambda x: int(UiMod.create(x['ref']).id) % 10,
                unit_types=[uimod_type])

            uimod_pid = uimod_pid or UiMod.create(root_info['ref']).pid
            for candidate in range(10):
                if candidate not in used_uimod_rel_ids:
                    uimod_id = 10 * int(uimod_pid) + candidate
                    eventlog.log_d(f'done: pid:{uimod_pid}, type:{uimod_type}, id:{uimod_id}')
                    return str(uimod_pid), uimod_type, str(uimod_id)

        raise NotImplementedError

    @classmethod
    def validate_term(cls, term_pid: str) -> bool:
        root_info = cls.get_info()
        term_uimod = UiMod.create_term(term_pid)
        term_info = cls.find_unit_r(root_info, unit_ref=term_uimod.ref)
        return bool(term_info)

    @classmethod
    def get_term_info(cls, term_uuid: str) -> dict | None:
        root_info = cls.get_info()
        terms_info = cls.map_units_r(root_info, unit_types=[uitype.UI_TYPE_TERMINAL])

        for term_info in terms_info:
            term_uimod = UiMod.create(term_info['ref'])
            term_uiconf = term_uimod.load_conf()
            if term_uuid == term_uiconf['ro']['uuid']:
                return term_info

    @classmethod
    def get_term_pid(cls, term_uuid: str) -> str | None:
        term_info = cls.get_term_info(term_uuid)
        if term_info:
            term_uimod = UiMod.create(term_info['ref'])
            return term_uimod.pid

    @classmethod
    def get_term_event_types(cls, term_info: dict) -> list[str]:
        det_types = [
            uitype.UI_TYPE_DET_TRANSIT,
            uitype.UI_TYPE_DET_PARKING,
            uitype.UI_TYPE_DET_WRONG_WAY,
            uitype.UI_TYPE_DET_REDLIGHT,
            uitype.UI_TYPE_DET_WANTED,
            uitype.UI_TYPE_DET_SECTION_SPEED,
            uitype.UI_TYPE_DET_BUS_LANE,
            uitype.UI_TYPE_DET_SEATBELT
        ]

        if cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_RADAR):
            det_types.append(uitype.UI_TYPE_DET_SPOT_SPEED)

        return det_types

    @classmethod
    def get_raterm_pid(cls) -> str | None:
        root_info = cls.get_info()
        radar_info = cls.find_unit_r(root_info, unit_type=uitype.UI_TYPE_RADAR)
        if radar_info:
            return UiMod.create(radar_info['ref']).pid

    @classmethod
    def add_term(cls, ctrl_uiref: str,
                 factory_info: dict,
                 factory_links: list[tuple[str, str]],
                 internal: bool = False) -> tuple[dict, dict]:

        root_info, root_links = cls.get_info_links()

        uimod_pid, _, _ = cls.new_uimod_pid_type_id(root_info, uitype.UI_TYPE_TERMINAL)

        def reduce_fn(acc: dict[str, str], old_uiref: str) -> dict[str, str]:
            uimod_type = UiMod.create(old_uiref).type
            pid_type_id = cls.new_uimod_pid_type_id(factory_info, uimod_type, uimod_pid)
            acc[old_uiref] = UiMod(*pid_type_id).ref
            return acc

        old_factory_refs = cls.find_unit_refs_r(factory_info)
        old_to_new_factory_refs = reduce(reduce_fn, old_factory_refs, dict())

        # links
        for factory_link in factory_links:
            factory_link[0] = old_to_new_factory_refs.get(factory_link[0], factory_link[0])
            factory_link[1] = old_to_new_factory_refs.get(factory_link[1], factory_link[1])

        root_links.extend(factory_links)

        # conf
        def replace_unit_ref_fn(unit_info: dict) -> None:
            unit_info['ref'] = old_to_new_factory_refs[unit_info['ref']]

        term_info = factory_info
        term_info['internal'] = internal
        cls.map_units_r(term_info, replace_unit_ref_fn)

        ctrl_info = cls.find_unit_r(root_info, unit_ref=ctrl_uiref)
        ctrl_info['units'].append(term_info)

        def find_link(links_list: list[tuple[str, str]], ref1: str, ref2: str) -> bool:
            for link in links_list:
                if ref1 in link and ref2 in link:
                    return True

            return False

        tr_uirefs = cls.find_unit_refs_r(root_info, unit_types=[uitype.UI_TYPE_TRAFFIC_RULES])
        iit_ca_uirefs = cls.find_unit_refs_r(root_info, unit_types=[uitype.UI_TYPE_IIT_CA])
        iit_ca_eu_uirefs = cls.find_unit_refs_r(root_info, unit_types=[uitype.UI_TYPE_IIT_CA_END_USER])

        for tr_uiref in tr_uirefs:
            for iit_ca_uiref in iit_ca_uirefs:
                if not find_link(root_links, tr_uiref, iit_ca_uiref):
                    root_links.append((tr_uiref, iit_ca_uiref))

            for iit_ca_eu_uiref in iit_ca_eu_uirefs:
                if not find_link(root_links, tr_uiref, iit_ca_eu_uiref):
                    root_links.append((tr_uiref, iit_ca_eu_uiref))

        cls.set_info_links(root_info, root_links)
        return root_info, term_info

    @classmethod
    def del_unit_r(cls, unit_ref: str) -> tuple[dict, set[str]]:
        """
        Unlinks and removes a unit and its sub-units from the root_info tree.

        :param unit_ref:
            The ref of the unit to be removed.
        """
        root_info, root_links = cls.get_info_links()

        # find the unit to be removed and its (the most nested) parent.
        unit_info, parent_info = cls.find_unit_parent_r(root_info, unit_ref=unit_ref)
        if parent_info is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        # define the ref of the unit and the refs of its sub-units.
        del_unit_refs = set[str](cls.find_unit_refs_r(unit_info))
        assert len(del_unit_refs) > 0

        # unlink unit and its sub-units.
        unlinked_unit_refs = set[str]()
        for del_unit_ref in del_unit_refs:
            for link in root_links:
                if del_unit_ref in link:
                    unlinked_unit_refs.update(link)

        for del_unit_ref in del_unit_refs:
            root_links = [link for link in root_links if del_unit_ref not in link]

        # define the unit refs whose links have been changed.
        # the set does not contain refs of the deleted units.
        changed_unit_refs = unlinked_unit_refs - del_unit_refs

        # remove unit (and its sub-units) from parent
        parent_info['units'] = [x for x in parent_info['units'] if x['ref'] != unit_ref]

        cls.set_info_links(root_info, root_links)
        return root_info, changed_unit_refs

    @classmethod
    def add_detector(cls, det_uimod_pid: str, det_uimod_type: str, det_name: str) -> tuple[str, dict] | None:
        root_info, root_links = cls.get_info_links()

        term_uimod = UiMod.create_term(det_uimod_pid)
        term_info = cls.find_unit_r(root_info, unit_ref=term_uimod.ref)
        if term_info is None:
            eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_TERMINAL})')
            raise NotFoundError

        rules_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_TRAFFIC_RULES)
        if rules_info is None:
            eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_TRAFFIC_RULES})')
            raise NotFoundError

        _, _, det_uimod_id = cls.new_uimod_pid_type_id(rules_info, det_uimod_type)
        det_uimod = UiMod(det_uimod_pid, det_uimod_type, det_uimod_id)
        rules_info['units'].append({
            'ref': det_uimod.ref,
            'type': det_uimod.type,
            'name': det_name,
            'role': uitype.UI_ROLE_DETECTOR,
            'internal': False
        })

        if det_uimod_type == uitype.UI_TYPE_DET_SPOT_SPEED:
            http_streaming_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_HTTP_STREAMING)
            if http_streaming_info is None:
                eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_HTTP_STREAMING})')
            else:
                root_links.append((det_uimod.ref, http_streaming_info['ref']))

            rtsp_streaming_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_RTSP_STREAMING)
            if rtsp_streaming_info is None:
                eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_RTSP_STREAMING})')
            else:
                root_links.append((det_uimod.ref, rtsp_streaming_info['ref']))

        elif det_uimod_type in [uitype.UI_TYPE_DET_REDLIGHT, uitype.UI_TYPE_DET_RAILROAD]:
            http_streaming_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_HTTP_STREAMING)
            if http_streaming_info is None:
                eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_HTTP_STREAMING})')
            else:
                root_links.append((det_uimod.ref, http_streaming_info['ref']))

            rtsp_streaming_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_RTSP_STREAMING)
            if rtsp_streaming_info is None:
                eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_RTSP_STREAMING})')
            else:
                root_links.append((det_uimod.ref, rtsp_streaming_info['ref']))

            traffic_light_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_TRAFFIC_LIGHT)
            if traffic_light_info is None:
                eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_TRAFFIC_LIGHT})')
            else:
                root_links.append((det_uimod.ref, traffic_light_info['ref']))

            rtsp_cam_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_RTSP_CAMERA)
            if rtsp_cam_info is None:
                eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_RTSP_CAMERA})')
            else:
                root_links.append((det_uimod.ref, rtsp_cam_info['ref']))

            mv_cam_info = cls.find_unit_r(term_info, unit_type=uitype.UI_TYPE_MV_CAMERA)
            if mv_cam_info is None:
                eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_MV_CAMERA})')
            else:
                root_links.append((det_uimod.ref, mv_cam_info['ref']))

        root_links.append((det_uimod.ref, rules_info['ref']))
        cls.set_info_links(root_info, root_links)
        return det_uimod_id, root_info

    @classmethod
    def add_service(cls, service_uimod_pid: str, service_uimod_type: str, service_name: str,
                    parent_type: str, types_to_link: list[str]) -> tuple[str, dict] | None:

        root_info, root_links = cls.get_info_links()

        term_uimod = UiMod.create_term(service_uimod_pid)
        term_info = cls.find_unit_r(root_info, unit_ref=term_uimod.ref)
        if term_info is None:
            eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_TERMINAL})')
            raise NotFoundError

        parent_info = cls.find_unit_r(term_info, unit_type=parent_type)
        if parent_info is None:
            eventlog.log_d(f'fail: NotFoundError ({parent_type})')
            raise NotFoundError

        _, _, service_uimod_id = cls.new_uimod_pid_type_id(parent_info, service_uimod_type)
        service_uimod = UiMod(service_uimod_pid, service_uimod_type, service_uimod_id)

        if parent_info.get('units') is None:
            parent_info['units'] = []

        parent_info['units'].append({
            'ref': service_uimod.ref,
            'type': service_uimod.type,
            'name': service_name,
            'role': uitype.UI_ROLE_SERVICE,
            'internal': False
        })

        for type_to_link in types_to_link:
            unit_to_link_info = cls.find_unit_r(term_info, unit_type=type_to_link)
            root_links.append((service_uimod.ref, unit_to_link_info['ref']))

        cls.set_info_links(root_info, root_links)
        return service_uimod_id, root_info

    @classmethod
    def add_controller_service(cls, service_uimod_type: str, service_name: str, types_to_link: list[str],
                               controller_id: str = '0') -> tuple[str, dict] | None:

        root_info, root_links = cls.get_info_links()

        controller_uimod = UiMod.create_controller(controller_id)
        controller_info = cls.find_unit_r(root_info, unit_ref=controller_uimod.ref)
        if controller_info is None:
            eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_TERMINAL})')
            raise NotFoundError

        _, _, service_uimod_id = cls.new_uimod_pid_type_id(controller_info, service_uimod_type)
        service_uimod = UiMod('0', service_uimod_type, service_uimod_id)

        if controller_info.get('units') is None:
            controller_info['units'] = []

        controller_info['units'].append({
            'ref': service_uimod.ref,
            'type': service_uimod.type,
            'name': service_name,
            'role': uitype.UI_ROLE_SERVICE,
            'internal': False
        })

        for type_to_link in types_to_link:
            unit_to_link_info = cls.find_unit_r(controller_info, unit_type=type_to_link)
            root_links.append((service_uimod.ref, unit_to_link_info['ref']))

        cls.set_info_links(root_info, root_links)
        return service_uimod_id, root_info

    @classmethod
    def get_ext_root_info(cls, terms_info: dict | None = None) -> dict:

        if terms_info is None:
            terms_info = cls.get_info()

        return terms_info
        #
        # terms_info.sort(key=lambda x: x['id'])
        #
        # for term_info in terms_info:
        #     ui_pid = term_info['id']
        #     modules_info = term_info['modules']
        #
        #     try:
        #         DgStats.test_cam_status(ui_pid)
        #         cam_error = False
        #     except CamServiceError:
        #         cam_error = True
        #
        #     # set pipeline status
        #     cam_dep_types = [
        #         uitype.UI_TYPE_MV_CAMERA,
        #         uitype.UI_TYPE_RTSP_CAMERA,
        #         uitype.UI_TYPE_LPR,
        #         uitype.UI_TYPE_ROADWAY,
        #         uitype.UI_TYPE_HTTP_STREAMING
        #     ]
        #
        #     for module_info in modules_info:
        #         if module_info['type'] in cam_dep_types:
        #             module_info['error'] = cam_error
        #         else:
        #             module_info['error'] = False
        #
        #     term_uimod = UiMod(ui_pid, uitype.UI_TYPE_TERMINAL, str(10 * int(ui_pid)))
        #     term_info['limits'] = {
        #         'detectors': get_term_detectors_types(term_uimod.ref)
        #     }
        #
        # return terms_info

        # ctrl_info = InfoUiconf.find_unit_r(root_info, unit_type=uitype.UI_TYPE_CONTROLLER)
        # ctrl_units = ctrl_info['units']
        #
        # for unit_info in ctrl_units:
        #     if unit_info['type'] != 'terminal':
        #         continue
        #
        #     unit_info['factory'] = [
        #         {
        #             'role': 'detector',
        #             'type': 'parking'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'wrong-way'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'red-light'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'wanted'
        #         },
        #         {
        #             'role': 'detector',
        #             'type': 'section-speed'
        #         }
        #     ]
        #
        #     term_units_info = unit_info['units']
        #     ra_units_info = filter(lambda subunit_jnode: subunit_jnode['type'] == 'radar', term_units_info)
        #     if len(ra_units_info) > 0:
        #         unit_info['factory'].append({
        #             'role': 'detector',
        #             'type': 'spot-speed'
        #         })


# noinspection PyClassHasNoInit
class CamDgStats:
    def __init__(self, jnode_path: str):
        self._jnode = fs.load_json3(jnode_path, default=None)

    @classmethod
    def create(cls, term_pid: str) -> Self:
        ui_id = '0'
        jnode_path = fs.join_path(f'/var/spool/edge/pipeline-{term_pid}/zabbix/vs{ui_id}')
        return cls(jnode_path)

    def _steam_active(self) -> bool:
        stats_jnode = self._jnode
        if stats_jnode:
            stream_jnode = stats_jnode.get('stream')
            if stream_jnode is not None:
                # old format
                return stream_jnode['error'] == 0
            else:
                units_jnode: list[dict] = stats_jnode.get('units') or list()
                bus_streamers = [x for x in units_jnode if x.get('type') == 'bus-streamer']
                max_endpoints = 0
                max_jnode = dict()
                for bus_streamer in bus_streamers:
                    stat_jnode = bus_streamer.get('stat') or dict()
                    if stat_jnode.get('endpoints', 0) > max_endpoints:
                        max_jnode = stat_jnode

                return max_jnode.get('active', False)

        return False

    def test_availability(self) -> None:
        cam_status = self._steam_active()
        if not cam_status:
            eventlog.log_w('fail: CamServiceError (cam-status:<none>)')
            raise CamServiceError

    def get_stream_wh(self) -> tuple[int, int] | None:

        stats_jnode = self._jnode
        if stats_jnode:
            stream_jnode = stats_jnode.get('stream')
            if stream_jnode is not None:
                # old format
                return stream_jnode.get('wh')
            else:
                units_jnode: list[dict] = stats_jnode.get('units') or list()
                bus_streamers = [x for x in units_jnode if x.get('type') == 'bus-streamer']
                max_endpoints = 0
                max_jnode = dict()
                for bus_streamer in bus_streamers:
                    stat_jnode = bus_streamer.get('stat') or dict()
                    if stat_jnode.get('endpoints', 0) > max_endpoints:
                        max_jnode = stat_jnode

                return max_jnode.get('wh')

    def get_stats(self, term_pid: str) -> dict:
        term_uimod = UiMod.create_term(term_pid)
        term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
        stream_stats = {
            'stream': f'api/v2/origins/media/{term_uuid}/live/stream'
        }

        cam_stream_wh = self.get_stream_wh()
        if cam_stream_wh:
            stream_stats['resolution'] = cam_stream_wh

        return {
            'media': [stream_stats]
        }


class CamUiStats:
    def __init__(self, intrinsic_jnode: dict | None) -> None:
        self._intrinsic_jnode = intrinsic_jnode

    @classmethod
    def create(cls, term_pid: str) -> Self:
        mv_cam_uimod = UiMod(term_pid, uitype.UI_TYPE_MV_CAMERA, '0')
        rtsp_cam_uimod = UiMod(term_pid, uitype.UI_TYPE_RTSP_CAMERA, '0')

        mv_jnode = mv_cam_uimod.load_conf(default=None) or dict()

        mv_intr = mv_jnode.get('ro', {}).get('intrinsic')
        if mv_intr:
            return cls(mv_intr)
        else:
            rtsp_jnode = rtsp_cam_uimod.load_conf(default=None) or dict()
            rtsp_intr = rtsp_jnode.get('ro', {}).get('intrinsic') \
                or rtsp_jnode.get('rw', {}).get('conf', {}).get('intrinsic')

            if rtsp_intr:
                return cls(rtsp_intr)

        return cls(None)

    def get_sensorwpx_sensorhpx_focalpx(self) -> tuple[float, float, float] | None:

        if self._intrinsic_jnode:
            sensor_w_px, sensor_h_px = self._intrinsic_jnode['frustum']['sensor-size']
            focus_dist_px = self._intrinsic_jnode.get('calibration', dict()).get('focal-dist')
            if focus_dist_px is None:
                focus_dist_mm = self._intrinsic_jnode['frustum']['focal-dist']
                pixel_size_um = self._intrinsic_jnode['frustum']['pixel-size']
                focus_dist_px = focus_dist_mm / pixel_size_um * 1000

            return sensor_w_px, sensor_h_px, focus_dist_px


def get_terminal_stats(term_pid: str) -> dict:
    root_info = RootInfoUiconf.get_info()
    term_uimod = UiMod.create_term(term_pid)
    term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
    video_wh = CamDgStats.create(term_pid).get_stream_wh()

    stats_jnode = {
        'media': []
    }

    term_info = RootInfoUiconf.find_unit_r(root_info, unit_ref=term_uimod.ref)
    if term_info is None:
        eventlog.log_d(f'fail: NotFoundError ({uitype.UI_TYPE_TERMINAL})')
        raise NotFoundError

    http_streaming_info = RootInfoUiconf.find_unit_r(term_info, unit_type=uitype.UI_TYPE_HTTP_STREAMING)
    if http_streaming_info is not None:
        http_server_uimod = UiMod.create(http_streaming_info['ref'])
        http_server_uiconf = http_server_uimod.load_conf()
        http_server_uiconf_pub = http_server_uiconf['rw']
        http_enabled = http_server_uiconf_pub.get('enabled', False)
        if http_enabled:
            http_stats = {
                'stream': f'api/v2/origins/media/{term_uuid}/live/stream'
            }
            if video_wh:
                http_stats['resolution'] = video_wh

            stats_jnode['media'].append(http_stats)

    rtsp_streaming_info = RootInfoUiconf.find_unit_r(term_info, unit_type=uitype.UI_TYPE_RTSP_STREAMING)
    if rtsp_streaming_info is not None:
        rtsp_server_uimod = UiMod.create(rtsp_streaming_info['ref'])
        rtsp_server_uiconf = rtsp_server_uimod.load_conf()
        rtsp_server_uiconf_pub = rtsp_server_uiconf['rw']
        rtsp_enabled = rtsp_server_uiconf_pub.get('enabled', False)
        if rtsp_enabled:
            rtsp_stats = {
                'stream': rtsp_server_uiconf_pub.get('streaming', {}).get('url', 'try again later')  # TODO
            }
            stats_jnode['media'].append(rtsp_stats)

    return stats_jnode


def get_detector_limits_stats(
        term_pid: str,
        vehicle: bool = True,
        scene: bool = True,
        lanes: bool = False,
        classes: bool = False) -> dict:

    root_info = RootInfoUiconf.get_info()

    def store_factory(es_info: dict) -> dict:
        return {
            'id': es_info['ref'].split('.')[-1],  # TODO: for current ui
            'ref': es_info['ref'],
            'name': es_info.get('name')
        }

    term_uimod = UiMod.create_term(term_pid)
    term_uuid = TerminalUiconf(term_uimod.conf_path).get_uuid()
    video_wh = CamDgStats.create(term_pid).get_stream_wh()
    video_w, video_h = (640, 480) if video_wh is None else video_wh

    stats_jnode = {
        'media': [{
            'resolution': [video_w, video_h],
            'stream': f'api/v2/origins/media/{term_uuid}/live/stream'
        }]
    }

    limits_jnode = {
        'images': dict(),
        'stores': InfoUiconf.map_units_r(root_info, store_factory, unit_types=[uitype.UI_TYPE_EVENTSTORE])
    }

    if vehicle:
        limits_jnode['images']['vehicle'] = {'max-width': 640}

    if scene:
        limits_jnode['images']['scene'] = {'max-width': video_w}

    if lanes:
        roadway_uimod = UiMod(term_pid, uitype.UI_TYPE_ROADWAY, '0')  # TODO
        roadway_uiconf = roadway_uimod.load_conf()
        roadway_lanes = roadway_uiconf['rw']['args']['lanes']
        limits_jnode['lanes'] = [x['name'] for x in roadway_lanes]

    if classes:
        limits_jnode['classes'] = MMR_CLASSES

    return {
        'limits': limits_jnode,
        'stats': stats_jnode
    }


if __name__ == '__main__':
    pass
    # InfoUiconf._CONF_PATH = \
    #     '/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/root.json'
    # RootInfoUiconf._CONF_PATH = \
    #      '/home/user/Documents/BitBucket/edgeserver/_share/edge-python-tools/edge-python-tools/share/misc/root.json'
    #
    # root_info, links = RootInfoUiconf.get_info_links()
    # term_pid = '0'
    # term_uimod = UiMod.create_term(term_pid)
    #
    # term_info, parent_info = InfoUiconf.find_unit_parent(root_info, uitype.UI_TYPE_CONTROLLER, term_uimod.ref)
    #
    # def rm_link(unit_info):
    #     ref = unit_info['ref']
    #     print ref
    #     a = filter(lambda link: ref not in link, links)
    #     return a
    #
    # # reply = InfoUiconf.map_units(term_info, rm_link)
    # units_info = InfoUiconf.map_units(term_info, lambda x: x)
    # for unit_info in units_info:
    #     ref = unit_info['ref']
    #     links = filter(lambda link: ref not in link, links)
    #
    # import ujson
    # print ujson.dumps(links)
